package DTDDoc;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import com.wutka.dtd.DTD;
import com.wutka.dtd.DTDAttlist;
import com.wutka.dtd.DTDAttribute;
import com.wutka.dtd.DTDComment;
import com.wutka.dtd.DTDContainer;
import com.wutka.dtd.DTDElement;
import com.wutka.dtd.DTDEmpty;
import com.wutka.dtd.DTDEntity;
import com.wutka.dtd.DTDItem;
import com.wutka.dtd.DTDName;
import com.wutka.dtd.DTDPCData;
import com.wutka.dtd.DTDParser;

/** An object of this class represents a full DTD. That is, the result of
 * interpretation of a DTD file and all the other files it includes.
 * A DTD defines a bunch of elements and attributes (and small other things
 * as well). A DTD file can includes other DTD's file. Therefore some
 * pieces, although defined in some other files, are defined in the
 * DTD represented by <code>this</code>. We call such pieces "external".
 *
 * One of the goals of this class is to improve over Mark Wutka's DTD class so
 * that one can know if an element/attribute is "external" or not.
 *
 * @author Stefan Champailler
 */

public class ExtendedDTD {

    private DefinitionsMap definitionsMap;

    /** Full path to the original DTD. */
    private File systemPath;

    /** The DTD, as parsed by Wutka's DTD parser. */
    private DTD dtd;

    /** Title of the DTD. See the @title tag.*/
    private String title = null;

    /** Encoding of the file. For example, ISO-8859-1. This is dictated
     *  by the <? ... encoding="" ?> processing instruction.
     */

    private String encoding = null;

    public String getEncoding() {
        return encoding;
    }

    /** Mapping between each element of the dTD and its parent (as long as
     *  the parent is defined in this DTD. */
    private Map parents = null;

    /** Returns the title of the DTD (see the @title tag).
     *  @return the title of the DTD. */
    public String getTitle() {
        return title;
    }

    /** Return all the DTDItems of the DTD (a transparent wrapper around
     *  the same function in Wutka's library).
     * @return The DTD Items of the DTD. */
    public Vector getItems() {
        return dtd.items;
    }

    /** Return all the DTDElements of the DTD (a transparent wrapper around
     *  the same function in Wutka's library).
     * @return A mapping from the name of the elements to their instance. */
    public Hashtable getElements() {
        return dtd.elements;
    }
    
    public Collection getElementsCollection() {
    	return dtd.elements.values();
    }
    
    public DTDElement getElementByName( String name) {
    	Object o = getElements().get( name);
    	if( o != null)
    		return (DTDElement) o;
    	else {
    		System.out.println("ERROR: Invalid name, or the " +
    				"element doesn't exist. name=" + name + ".");
    		return null;
    	}
    }
    
    public DTDElement getElementByName( DTDName name) {
    	return getElementByName( name.getValue());
    }
    
    /** Return all the DTDEntities of the DTD (a transparent wrapper around
     *  the same function in Wutka's library).
     * @return A mapping from the name of the entities to their instance. */
    public Hashtable getEntities() {
    	return dtd.entities;
    }

    /** Returns the root elements of the DTD, as defined by the \@root tag.
      * @return the root element. */
    public DTDElement getRootElement() {
        return dtd.rootElement;
    }

    /** Setter for the root elements of the DTD.
      * @param element The root element.*/
    public void setRootElement(DTDElement element) {
        dtd.rootElement = element;
    }

    /** Returns the full path to the original DTD.
     *  @return Full path to the original DTD.. */
    public File getSystemPath() {
        return systemPath;
    }

    /** Tells if an element is declared in a DTD outside the one represented in
     * <code>this</code>.
     *
     * @param element element to be checked for externality
     * @return true if external, false if not (or element not part of the DTD's
     *     elements. */

    public boolean isExternal(DTDElement element) {
        String elementPath =
            (String) definitionsMap.elementsLocations.get(element.name);
        return !elementPath.equals(getSystemPath().getPath());
    }

    /** Determine if an attributes list is defined in this DTD.
     *  @param attList attributes list to look for.
     *  @return true if the list is defined outside this DTD (therefore
     *      in another one. */

    public boolean isExternal(DTDAttlist attList) {
        return isExternal((DTDElement) getElements().get(attList.getName()));
    }

    /** Gives the file name of the DTD where a given element is defined.
     *
     * @param element element to be checked
     * @return the file name or <code>null</code> if the element is
     *     not part of the DTD. */

    public String getElementOrigin(DTDElement element) {
        String elementPath =
            (String) definitionsMap.elementsLocations.get(element.name);

        if (elementPath == null)
            System.out.println(
                "Requesting info about a non existing element : "
                    + element.name);
        return elementPath;
    }

    /** Build all the information needed to make a ExtendedDTD out of
     *  a "normal" DTD. The original DTD will be read with Mark Wutka's
     *  powerful DTD parser.
     *
     *  @param dtdFilePath Path to the original DTD. */

    public ExtendedDTD(File dtdFilePath) throws IOException {
        systemPath = dtdFilePath;

        if (!getSystemPath().canRead())
            throw new IOException(
                "Can't read "
                    + getSystemPath()
                    + ". Be prepared to get tons of errors !");

        definitionsMap = locateElements(getSystemPath().getPath());

        DTDParser dtdParser = new DTDParser(getSystemPath());
        dtd = dtdParser.parse(true);
        encoding = dtdParser.getDTDEncoding();

        setRootElement();
        attributesListMap = makeAttributesListsMap();
        parents = findParents(dtd);
        guessRoots( parents);
    }

    /** For a given DTD, this function will build a hashtable
     *  associating each of its elements (included those defined in
     *  its children) to the system-name of the DTDs in which those elements
     *  are defined. This function is necessary because Wutka's code
     *  doesn't do that correctly.
     *
     *  This function is VERY INEFFICIENT !
     *
     *  @param filePath where the DTD file is located.
     */

    private DefinitionsMap locateElements(String filePath) throws IOException {

        // Read the DTD
        // Very inefficient stuff here ...

        DTD dtd = null;
        File f = new File(filePath);

        try {
            DTDParser dtdParser = new DTDParser(f);
            dtd = dtdParser.parse(true);
        } catch (Exception ex) {
            System.out.println("locateElements():can't read " + filePath);
            return null;
        }

        // Analysing...

        DefinitionsMap defMap = new DefinitionsMap();

        // First, collect the position of elements defined in children DTDs.
        // Children DTD are enumerated in the entities.

        Enumeration items = dtd.entities.elements();
        while (items.hasMoreElements()) {

            DTDEntity entity = (DTDEntity) items.nextElement();

            // We're only interested in the *external DTD files*. Therefore
            // we have the following tests.

            String nextFile = null;

            if (entity.getExternalID() != null)
                nextFile = entity.getExternalID().getSystem();

            // We work only with files that are on the current filesystem
            // FIXME: break this limitation please !

            if (nextFile != null
                && !(nextFile.startsWith("http:")
                    || nextFile.startsWith("file:"))) {

                // FIXME: We're not sure if this is gonna work for files
                // that are placed in different directories. This depends
                // on the nature of externalID.system (which we don't know
                // for sure yet).

                nextFile =
                    filePath.substring(
                        0,
                        filePath.lastIndexOf(File.separatorChar) + 1)
                        + nextFile;

                // System.out.println( "///// diving in "+nextFile);
                defMap.merge(locateElements(nextFile));
            }
        }

        // Now that we know which elements are defined in the children,
        // we can find out those which are defined in the present DTD.
        // The idea is simple, an element is defined in this DTD (and not
        // in its children) if an only if it's not defined in its
        // children. Sometimes we work on obvious things...

        items = dtd.elements.elements();
        while (items.hasMoreElements()) {
            DTDElement element = (DTDElement) items.nextElement();

            if (!defMap.elementsLocations.containsKey(element.name))
                defMap.elementsLocations.put(element.name, filePath);
        }

        return defMap;
    }

    /** Extract the file name out of a full path. For example,
     *  for "/alpha/bravo/zulu", this function will return "zulu".
     *
     *  @param path The full path.
     *  @return The file name part of the path. */

    private String getFilename(String path) {

        // Warning ! lastIndexOf can return -1 !!!
        int li = path.lastIndexOf(File.separator) + 1;

        return path.substring(li);
    }

    /** Tag starting a root element defintion (@root) */
    private static final String ROOT_TAG = CommentParser.BEGIN_TAG + "root";

    /** Tag starting a DTD title defintion (@title) */
    private static final String TITLE_TAG = CommentParser.BEGIN_TAG + "title";

    /** This function will locate the "@root" element in the comments
      * in a given DTD.
      * With that information, it will determine the root of the
      * DTD and update the necessary member (i.e. rootElement).
      *
      * @return the value of the first "@root" doc-tag found.
      */

    private String findRootTagValue() {

        String rootName = null;
        Enumeration elements = getItems().elements();

        while (elements.hasMoreElements()) {

            Object obj = elements.nextElement();

            // Are we on a comment ?
            if (obj instanceof DTDComment) {

                DTDComment comment = (DTDComment) obj;

                // Since all the comments from the included files also
                // appear in the DTD, we have to make sure we only look
                // at the relevant one.

                // Then, we parse it... (we avoid a cast by using toString() !)
                CommentParser cp = new CommentParser(comment);
                //Hashtable p = cp.parseComment( comment.toString());
                String rn = cp.getUniqueTagValue(ROOT_TAG);

                // So, if we found the @root tag, it's cool but we have to
                // make sure that this one is related to the DTD and not one
                // of its children.

                if (rn != null) {
                    if (getElements().get(rn) != null) {
                        if (!isExternal((DTDElement) getElements().get(rn)))
                            rootName = rn;
                    } else
                        System.out.println(
                            "ERROR: The root element you "
                                + "specified with \""
                                + ROOT_TAG
                                + " "
                                + rn
                                + "\" "
                                + "doesn't exist in the DTD !");
                }

                // Now we figure out the title
                String t = cp.getUniqueTagValue(TITLE_TAG);
                if (t != null)
                    title = t;
            }
        }

        if (title == null) {
            // Title tag is not mandatory, therefore, I don't show a
            // warning anymore 
            //System.out.println("Didn't find any @title tag here !");
            title = getFilename(systemPath.getName());
        }
/*
        if (rootName == null)
            System.out.println(
                "Warning: Didn't find any valid @root tag here !");
*/
        return rootName;
    }

    /** This function will set the rootElement memeber in the dtd
      * according to the value to be found in the "@root" doc-tag. */

    private void setRootElement() {

        String rootName = findRootTagValue();

        // Did we find the @root doc-tag.
        if (rootName != null) {

            // Yes, make sure that this root really exists
            if (getElements().get(rootName) != null) {

                // It does ! Now we can update the DTD.
                setRootElement((DTDElement) getElements().get(rootName));

            } else

                // It doesn't. Therefore...
                System.out.println(
                    "setRootElement(): The provided root element ("
                        + ROOT_TAG
                        + " "
                        + rootName
                        + ") is nowhere to be found in the DTD!");

        }
    }

    /** Gives a unique id for an attribute in a DTD. The id construction
     *  relies on the fact that an attribute is always related to a element.
     *  That is each (element, attribute) pair is unique in a DTD.
     *  
     * @param attList The attributes list the attibute is part of.
     * @param attribute The attribute for which we seek the id.
     * @return The id of the attribute.
     */

    public String makeUniqueId( DTDAttlist attList, DTDAttribute attribute) {
    	// attList.name is the element's name.
        return attList.name + "_" + attribute.getName();
    }

    /** Gives a unique id for an attribute in a DTD.
     * 
     * @param element The element to which the attribute belongs.
     * @param attribute The attribute to get the id for.
     * @return The id of the attribute.
     * @see makeUniqueId( DTDAttlist attList, DTDAttribute attribute)
     */
    public String makeUniqueId( DTDElement element, DTDAttribute attribute) {
        return element.name + "_" + attribute.getName();
    }
    
    /** Gives a unique id for an element of a DTD.
     * 
     * @param element The element.
     * @return The id of the element.
     */

    public String makeUniqueId( DTDElement element) {
		return element.getName();		
	}
	
    private Hashtable attributesListMap;

    public DTDAttlist locateAttributesList(DTDAttribute attribute) {
        return (DTDAttlist) attributesListMap.get(attribute);
    }

    private Hashtable makeAttributesListsMap() {

        Hashtable hash = new Hashtable();

        for (int i = 0; i < getItems().size(); i++)
            if (getItems().get(i) instanceof DTDAttlist) {

                DTDAttlist attList = (DTDAttlist) getItems().get(i);

                for (int j = 0; j < attList.attributes.size(); j++)
                    hash.put(attList.attributes.get(j), attList);
            }

        return hash;
    }

    /** Finds the elements of a DTD which are "roots", that is which do not
     * have parent.
     * 
     * @param parents A map mapping each element onto a set containing its
     *     children (or null if the set is empty).
     * @return The elements which are root. An empty collection if none are
     *     (might be possible with an empty DTD).
     */
    
    private Collection guessRoots( Map parents) {
    
    	HashSet roots = new HashSet();
    	
    	Iterator i = parents.entrySet().iterator(); 
    	while( i.hasNext()) {
    		Map.Entry entry = (Map.Entry) i.next();
    		if( entry.getValue() == null) {
    			roots.add( entry.getKey());
    			//System.out.println("ROOT !!! "+entry.getKey());
    		}
    	}
    	
    	return roots;
    }
    
    /** Builds an map associating each descendant of an element to its parent
     *  (the element itself :)). All the elements of the given DTD are
     *  examined. 
     *
     *  @param dtd A DTD where to look for the elements
     *  @return A Hashtable associating the names (string) of each element
     *      to a List of its parents (most of the time, only one parent). */

    private Map findParents(DTD dtd) {
        HashMap parents = new HashMap();

        Iterator i = dtd.elements.values().iterator();
        
        // We scan all the elements of the DTD
        while(i.hasNext()) {

            DTDElement element = (DTDElement) i.next();

            // We make sure that each element appears in the map.
            
            String k = element.getName();
            
            if( !parents.containsKey( k))
            	parents.put( k, null);
            
            // Now we set each element's child parent to the element.
            DTDItem content = element.getContent();
            
            // Not all elements have children (for example, an
            // element can be EMPTY>
            Set children = collectDeclaredChildren( element);
            
            if( children != null)
            	for ( Iterator j = children.iterator(); j.hasNext();) {
            		Object o = j.next();
            		
            		if (o instanceof DTDName) {
            			
            			String cName = ((DTDName) o).getValue();
            			
            			// Each child of e has e as parent.
            			// Except that an element can't be his
            			// own parent.
            			
            			if( !element.getName().equals( cName)) {
            				HashSet l = (HashSet) parents.get(cName);
            				
            				if (l == null)
            					l = new HashSet();
            				
            				l.add(element);
            				
            				parents.put(cName, l);
            			}
            			// System.out.println(children[j].toString()+"("+ ((DTDName)children[j]).getValue() +")" + " --> "+ e+ "("+((DTDElement)e).getName()+")");
            		}
            	}
        }
        
        return parents;
    }

    /** Gets the parents of an element.
     * 
     * @param sonName The name of the element to look for.
     * @return Null if the element doesn't exist or it has no parent.
     */
    
    public Set getParents(String sonName) {
        return (HashSet) parents.get(sonName);
    }
    
    /** Check if an element is root, that is, if it has no parent.
     * 
     * @param element The element to check.
     * @return True if its is root, false else.
     */
    
    public boolean isRoot( DTDElement element) {
    	Set parents = getParents( element.getName());
    	return parents == null || parents.size() == 0;
    }

    /** Collects the elements that appear in the definition of
     * a given element (its "children").
     * 
     * <p>Special definitions such as ANY, EMPTY and PCDATA are ignored.</p>
     * 
     * @param item The element.
     * @return The set of children or null if there ain't any child.
     */
    
    public Set collectDeclaredChildren( DTDElement element) {
    	return collectDeclaredChildrenHelper( element.getContent());
    }
    
    private Set collectDeclaredChildrenHelper( DTDItem item) {

    	if (item instanceof DTDContainer) {
    		
    		Enumeration items = ((DTDContainer) item).getItemsVec().elements();
    		
    		Set children = null;
    		
    		while (items.hasMoreElements()) {
    			Set s = collectDeclaredChildrenHelper( (DTDItem)
    					items.nextElement());
    			if (s != null) {
    				if( children == null) children = new HashSet();
    				children.addAll(s);
    			}
    		}
    		
    		return children;
    		
    	} else if (item instanceof DTDName)
    		
    		return Collections.singleton( item);
    		
    	else
    		// ANY, PCDATA and EMPTY are not considered as children, they
    		// are not named.
    		return null;
    }
    
}
